//Copyright (c) 2004, Michael Schubert
//All rights reserved.
//Parts of this code are based on
//NEncode (C) 2003 DPeshev.
//See copyright notice in AudioPluginmanager.h

#undef NDEBUG
#pragma comment(linker,"/ignore:4089")
#include <afx.h>
#include <atlbase.h>
#include "NeroAPI\AudioBase.h"
#include "AudioPluginManager.h"

#define NAACVersion "1.20"

CString		GetDLLDir();
void			PrintUsage();
BOOL			FileExists(CString csFile);

int main(int argc, char* argv[])
{
	if (argc < 2)
	{
		PrintUsage();
		return -1;
	}

	CString csDLLDir = GetDLLDir();
	if (csDLLDir != "")
	{
		if (csDLLDir.Right(1) != "\\")
			csDLLDir += "\\";
	}

	//Parse the command line
	CString csCurrentArg = "";
	DWORD dwDefVbrMode = 7;
	DWORD dwDefBitRate = 128;
	//Mode
	csCurrentArg = argv[1];
	csCurrentArg.MakeLower();
	if (csCurrentArg == "-tape")
		dwDefVbrMode = 3;
	else if (csCurrentArg == "-radio")
		dwDefVbrMode = 4;
	else if (csCurrentArg == "-internet")
		dwDefVbrMode = 5;
	else if (csCurrentArg == "-streaming")
		dwDefVbrMode = 6;
	else if (csCurrentArg == "-normal")
		dwDefVbrMode = 7;
	else if (csCurrentArg == "-extreme")
		dwDefVbrMode = 8;
	else if (csCurrentArg == "-audiophile")
		dwDefVbrMode = 9;
	else if (csCurrentArg == "-transcode")
		dwDefVbrMode = 10;
	else if (csCurrentArg.Left(4) == "-cbr")
	{
		dwDefVbrMode = 0;
		dwDefBitRate = atoi((LPCTSTR)(csCurrentArg.Right(csCurrentArg.GetLength() - 4)));
		switch (dwDefBitRate)
		{
			case 16:
			case 20:
			case 24:
			case 28:
			case 32:
			case 40:
			case 48:
			case 56:
			case 64:
			case 80:
			case 96:
			case 112:
			case 128:
			case 160:
			case 192:
			case 224:
			case 256:
			case 320:
			case 388:
			case 448:
				break;
			default:
				printf("\nPlease use one of the predefined CBR bitrates.\n");
				return -1;
		}
	}
	else
	{
		printf("\nInvalid VBR/CBR mode.\n");
		return -1;
	}

	//Switches and files
	DWORD dwDefCodecQuality  = 1;
	DWORD dwDefDownmix  = 0;
	DWORD dwDefExportAAC  = 0;
	DWORD dwDefHintForStreaming  = 0;
	DWORD dwDefUsePNS  = 0;
	DWORD dwDefMuxIntoVideo  = 0;
	DWORD dwDefAacProfile = 4; //LC = 4, HE = 5 
	CString csWAVFile = "";
	CString csAACFile = "";
	CString csMsg = "";
	BOOL bAACProfileOverride = FALSE;

	for (int iArg = 2; iArg < argc; iArg++)
	{
		csCurrentArg = argv[iArg];
		csCurrentArg.MakeLower();

		if (csCurrentArg == "-qh")
			dwDefCodecQuality  = 1;
		else if (csCurrentArg == "-qf")
			dwDefCodecQuality  = 0;
		else if (csCurrentArg == "-lc")
		{
			dwDefAacProfile = 4;
			bAACProfileOverride = TRUE;
		}
		else if (csCurrentArg == "-he")
		{
			dwDefAacProfile = 5;
			bAACProfileOverride = TRUE;
		}
		else if (csCurrentArg == "-mono")
			dwDefDownmix = 1;
		else if (csCurrentArg == "-exportiso")
			dwDefExportAAC = 1;
		else if (csCurrentArg == "-hinttrack")
			dwDefHintForStreaming = 1;
		else if (csCurrentArg == "-pns")
			dwDefUsePNS = 1;
		else if (csCurrentArg == "-muxvideo")
			dwDefMuxIntoVideo = 1;
		else //no more switches, must be the wav file or an invalid switch
		{
			if (csCurrentArg.Right(4) == ".wav") //wav
			{
				csWAVFile = argv[iArg];
				if (FileExists(csWAVFile) == FALSE)
				{
					csMsg.Format("\nFile not found: %s\n", csWAVFile);
					printf(csMsg);
					return -1;
				}
				break;	
			}
			else //must be an invalid switch
			{
				csMsg.Format("\nInvalid switch: %s\n", csCurrentArg);
				printf(csMsg);
				return -1;
			}
		}
	}

	if (bAACProfileOverride == FALSE)
	{
		if (dwDefVbrMode == 0) //CBR
		{
			if (dwDefBitRate > 96)
				dwDefAacProfile = 4; //LC
			else
				dwDefAacProfile = 5; //HE
		}
		else //VBR
		{
			if ((dwDefVbrMode >= 3) && (dwDefVbrMode <= 6))
				dwDefAacProfile = 5; //HE
			else
				dwDefAacProfile = 4; //LC	
		}
	}

	if (dwDefAacProfile == 4)
		dwDefUsePNS = 0;

	csAACFile = argv[iArg + 1];

	//Update registry
	HKEY hResult = 0;
	DWORD dwDisposition = 0;
	BOOL bRegError = FALSE;
	DWORD dwDummy = 0;
	if (RegCreateKeyEx(HKEY_CURRENT_USER,
										 "Software\\ahead\\Nero - Burning Rom\\Nero Digital Audio (HE-AAC)",
										 0,
										 "",
										 REG_OPTION_NON_VOLATILE,
										 KEY_ALL_ACCESS,
										 NULL,
										 &hResult,
										 &dwDisposition) != ERROR_SUCCESS)
	{
		printf("\nCannot access registry.\n");
		return -1;
	}

	if (RegSetValueEx(hResult, "DefBitRate", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefBitRate, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefVbrMode", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefVbrMode, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefCodecQuality", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefCodecQuality, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefAacProfile", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefAacProfile, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefUsePNS", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefUsePNS, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefHintForStreaming", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefHintForStreaming, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefOptimize", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDummy, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefDownmix", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefDownmix, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefMuxIntoVideo", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefMuxIntoVideo, 4) != ERROR_SUCCESS)
		bRegError = TRUE;
	if (RegSetValueEx(hResult, "DefExportAAC", 0, REG_DWORD, (LPBYTE)(DWORD)&dwDefExportAAC, 4) != ERROR_SUCCESS)
		bRegError = TRUE;

	RegCloseKey(hResult);
	RegFlushKey(HKEY_CURRENT_USER);

	if (bRegError == TRUE)
	{
		printf("\nCannot access registry.\n");
		return -1;
	}

	CString csDLL = "";
	HINSTANCE hWavDLL, hAacDLL, hAacEncDLL;
	csDLL = csDLLDir + "aacenc32.dll";
	hAacEncDLL = LoadLibrary((LPCTSTR)csDLL);
	csDLL = csDLLDir + "aac.dll";
	hAacDLL = LoadLibrary((LPCTSTR)csDLL);
	csDLL = csDLLDir + "wav.dll";
	hWavDLL = LoadLibrary((LPCTSTR)csDLL);

	if ((hAacEncDLL == 0) || (hAacDLL == 0) || (hWavDLL == 0))
	{
		printf("\nCannot find Nero AAC DLL's\nor\nNero is not installed.\n");
		return -1;
	}

	int nRetCode = 0;
	CAudioPluginMgr pluginMgr;
	IStatus *pIStatus;
	typedef bool (* LPNERO_PLUGIN_GetPrimaryAudioObject)(IAudioComponent**);
	typedef bool (* LPNERO_PLUGIN_ReadyToFinish)();

	LPNERO_PLUGIN_GetPrimaryAudioObject wavGetPrimaryAudioObject,aacGetPrimaryAudioObject;
	LPNERO_PLUGIN_ReadyToFinish wavReadyToFinish,aacReadyToFinish;
	IAudioComponent *pWavIA,*pAacIA;

	wavGetPrimaryAudioObject = (LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress(hWavDLL,"NERO_PLUGIN_GetPrimaryAudioObject");
	assert(wavGetPrimaryAudioObject);

	wavReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress(hWavDLL,"NERO_PLUGIN_ReadyToFinish");
	assert(wavReadyToFinish);

	aacGetPrimaryAudioObject = (LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress(hAacDLL,"NERO_PLUGIN_GetPrimaryAudioObject");
	assert(aacGetPrimaryAudioObject);

	aacReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress(hAacDLL,"NERO_PLUGIN_ReadyToFinish");
	assert(aacReadyToFinish);

	nRetCode = wavGetPrimaryAudioObject(&pWavIA); 
	assert(nRetCode && pWavIA->GetType() == ACT_ComponentEnumerator);

	nRetCode = aacGetPrimaryAudioObject(&pAacIA); 
	assert(nRetCode && pAacIA->GetType() == ACT_ComponentEnumerator);

	IConfigurable *pIC;
	IComponentEnum *pWavICU,*pAacICU; 
	nRetCode = pWavIA->QueryInterface(IID_IComponentEnum,(void **)&pWavICU);
	assert(!nRetCode && pWavICU && pWavICU->GetCount()>0);

	nRetCode = pAacIA->QueryInterface(IID_IComponentEnum,(void **)&pAacICU);
	assert(!nRetCode && pAacICU && pAacICU->GetCount()>1);

	IAudioComponent *pWavIAC,*pAacIAC;
	nRetCode = pWavICU->GetComponent(0,&pWavIAC);
	assert(nRetCode && pWavIAC && pWavIAC->GetType() == ACT_AudioURLSourceFactory);

	nRetCode = pAacICU->GetComponent(1,&pAacIAC);
	assert(nRetCode && pAacIAC && pAacIAC->GetType() == ACT_AudioURLTargetFactory);

	IURLAudioSourceFactory *pWavSrc;
	nRetCode = pWavIAC->QueryInterface(IID_IURLAudioSourceFactory, (void**)&pWavSrc);
	assert(!nRetCode && pWavSrc);

	IURLAudioTargetFactory *pAacDest;
	nRetCode = pAacIAC->QueryInterface(IID_IURLAudioTargetFactory, (void**)&pAacDest);
	assert(!nRetCode && pAacDest);

	nRetCode = pAacIAC->QueryInterface(IID_IConfigurable, (void**)&pIC);
	assert(!nRetCode && pIC);

	nRetCode = pAacIAC->Init(static_cast<IAudioPluginMgr*>(&pluginMgr), &pIStatus);
	assert(nRetCode && !pIStatus);

	IUnknown *pIUWav,*pIUAac;

	TCHAR tcSP[_MAX_PATH];
	assert(_tfullpath(tcSP, (LPCTSTR)csWAVFile, _MAX_PATH) != NULL);
	nRetCode = pWavSrc->Open(tcSP,&pIUWav);
	assert(nRetCode && pIUWav);

	IAudioRawReader *pWavIAURR;
	nRetCode = pIUWav->QueryInterface(IID_IAudioRawReader, (void**)&pWavIAURR);
	assert(!nRetCode && pWavIAURR);

	SWavFormat format = pWavIAURR->GetRawFormat();
	nRetCode = pAacDest->CreateURLAudioTarget(&pIUAac,format,&pIStatus);
	assert(nRetCode && !pIStatus);

	IAudioRawWriter *pAacIAURW;
	nRetCode = pIUAac->QueryInterface(IID_IAudioRawWriter, (void**)&pAacIAURW);
	assert(!nRetCode && pAacIAURW);

	IInfoWriter *pAacIW;
	nRetCode = pIUAac->QueryInterface(IID_IInfoWriter, (void**)&pAacIW);
	assert(!nRetCode && pAacIAURW);

	IURLHolder *pIURLH;
	nRetCode = pIUAac->QueryInterface(IID_IURLHolder, (void**)&pIURLH);
	assert(!nRetCode && pIURLH);

	TCHAR tcDP[_MAX_PATH];
	assert(_tfullpath(tcDP, (LPCTSTR)csAACFile, _MAX_PATH) != NULL);
	nRetCode = pIURLH->SetURL(tcDP,&pIStatus);
	assert(nRetCode && !pIStatus);

	IProcess *pAacIProcess,*pWavIProcess;

	nRetCode = pIUAac->QueryInterface(IID_IProcess, (void**)&pAacIProcess);
	assert(!nRetCode && pAacIProcess);

	nRetCode = pIUWav->QueryInterface(IID_IProcess, (void**)&pWavIProcess);
	assert(!nRetCode && pWavIProcess);

	nRetCode = pAacIProcess->Start(&pIStatus);
	assert(nRetCode && !pIStatus);

	nRetCode = pWavIProcess->Start(&pIStatus);
	assert(nRetCode && !pIStatus);

	BYTE pBuf[65536];
	int blockSize, blockSizeRead = 0;
	enum EAudioRawState state;
	ULONGLONG rawSize = pWavIAURR->GetRawLen();
	ULONGLONG size = rawSize;
	unsigned int uiSize = (unsigned int)size;
	unsigned int uiDone = 0;
	unsigned int uiPercentDone = 0;

	CString csPrt = "";
	csPrt.Format("\nEncoding %s\n", csWAVFile);
	printf((LPCTSTR)csPrt);
	
	unsigned int uiLength = 50;
	unsigned int i = 0;
	for (i = 1; i <= uiLength; i++)
		printf("%c", 177);
	for (i = 1; i <= uiLength; i++)
		printf("\b");

	unsigned int uiOldPercent = 0;
	while (size > 0)
	{
		blockSize = size > 65536 ? 65536 : (int)size;
		nRetCode = pWavIAURR->RawRead(pBuf, blockSize, &blockSizeRead, state, &pIStatus);
		assert(nRetCode && !pIStatus);
		nRetCode = pAacIAURW->RawWrite(pBuf, blockSizeRead, state, &pIStatus);
		assert(nRetCode && !pIStatus);
		size -= blockSize;
		uiDone = uiSize - (unsigned int)size;
		uiPercentDone = (unsigned int)(((double)uiDone / (double)uiSize) * uiLength);
		if (uiPercentDone > uiOldPercent)
		{
			for (unsigned int y = uiOldPercent; y < uiPercentDone; y++)
				printf("%c", 219);

			uiOldPercent = uiPercentDone;
		}
	}

	for (i = 1; i <= uiLength; i++)
		printf("\b");
	for (i = 1; i <= uiLength; i++)
		printf(" ");
	for (i = 1; i <= uiLength; i++)
		printf("\b");

	printf("Finished.\n");

	nRetCode = pAacIProcess->End(&pIStatus);
	assert(nRetCode && !pIStatus);

	nRetCode = pWavIProcess->End(&pIStatus);
	assert(nRetCode && !pIStatus);

	aacReadyToFinish();
	wavReadyToFinish();

	return 0;
}


void PrintUsage()
{
	printf("NAACEnc ");
	printf(NAACVersion);
	printf(" - (C) 2004 Michael Schubert\n");
	printf("Nero AAC encoder frontend based on NEncode (C) 2003 DPeshev.\n");
	printf("Requires Nero Burning Rom to be installed.\n\n");
	printf("Usage:\n");
	printf("NAACEnc <Mode> <Switches> <InFile.wav> <OutFile.mp4>\n\n");

	printf("Mode:\n\n");

	printf("  VBR:\n");
	printf("    -tape\n");
	printf("    -radio\n");
	printf("    -internet\n");
	printf("    -streaming\n");
	printf("    -normal\n");
	printf("    -extreme\n");
	printf("    -audiophile\n");
	printf("    -transcode\n\n");

	printf("  CBR:\n");
	printf("    -cbr<n> (n = 16,20,24,28,32,40,48,56,64,80,96,112,128,\n");
	printf("                 160,192,224,256,320,388 or 448 Kbps)\n\n");

	printf("Switches:\n\n");

	printf("  Encoder quality:\n");
	printf("    -qh (High) (Default)\n");
	printf("    -qf (Fast)\n\n");

	printf("  AAC profile:\n");
	printf("    -lc (Low complexity) (Default for CBR > 96 Kbps and VBR > 'streaming')\n");
	printf("    -he (High efficiency) (Default for CBR <= 96 Kbps and VBR <= 'streaming')\n\n");

	printf("  Misc:\n");
	printf("    -mono (Downmix to mono)\n");
	printf("    -exportiso (Export ISO 13818-7 AAC track)\n");
	printf("    -hinttrack (Create hint track for streaming server)\n");
	printf("    -pns (Enable PNS for very low bitrate)\n");
	printf("    -muxvideo (Mix audio into existing video .mp4 files)\n\n");

	printf("Examples:\n");
	printf("  NAACEnc -extreme test.wav test.mp4\n");
	printf("  NAACEnc -cbr96 test.wav test.mp4\n");

	return;
}


BOOL FileExists(CString csFile)
{
	DWORD dwAttrib = ::GetFileAttributes((LPCTSTR)csFile);
	if (dwAttrib == 0xFFFFFFFF)
		return FALSE;
	
	if (dwAttrib & FILE_ATTRIBUTE_DIRECTORY)
		return FALSE;

	return TRUE;
}


CString GetDLLDir()
{
	HKEY hKeyResult = 0;
	DWORD dwBytes = 0;
	char szDir[1024] = "";
	CString csDir = "";
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Software\\ahead\\shared", 0, KEY_READ, &hKeyResult) != ERROR_SUCCESS)
		return "";
	else
	{
		dwBytes = 500;
		RegQueryValueEx(hKeyResult, "AudioPlugIn", 0, 0, (LPBYTE)szDir, &dwBytes);
		csDir.Format("%s", szDir);
		RegCloseKey(hKeyResult);
	}

	return csDir;
}

